/*
 * Graphic license: https://openclipart.org/share
 */

#include "main.h"
#include "ui.h"
#include "user_callbacks.h"

#include <notification.h>
#include <storage.h>

static void win_delete_request_cb(void *data, Evas_Object *obj, void *event_info)
{
    ui_app_exit();
}

static void win_back_cb(void *data, Evas_Object *obj, void *event_info)
{
    appdata_s *ad = data;

    if (evas_object_visible_get(ad->xml_popup) == EINA_TRUE)
        evas_object_hide(ad->xml_popup);
    else
        elm_win_lower(ad->win);
}

static void _path_to_media_init(void *data)
{
    appdata_s *ad = data;
    int ret = storage_get_directory(ad->internal_storage_id, STORAGE_DIRECTORY_OTHERS, &ad->folder_path);
    if (STORAGE_ERROR_NONE != ret) {
        dlog_print(DLOG_DEBUG, LOG_TAG, "storage_get_directory() failed: %s ", ret);
        notification_status_message_post("Could not get folder path");
    }
    dlog_print(DLOG_DEBUG, LOG_TAG, "folder_path: %s ", ad->folder_path);
}

/*
 * Callback called after entering application id
 */
static void _pop_win_hide(void *data, Evas_Object *obj, void *event_info)
{
    appdata_s *ad = (appdata_s *)data;
    Evas_Object *entry = (Evas_Object *)evas_object_data_get(obj, "entry");
    Evas_Object *inwin = (Evas_Object *)evas_object_data_get(obj, "inwin");

    const char *app_id = elm_entry_entry_get(entry);

    /* Check whether app id is not empty */

    if (strlen(app_id) > 0) {
        strcpy(ad->app_id, app_id);
        evas_object_hide(inwin);

        /*
         * Create UI elements
         */
        _create_box(ad);
        _create_spinner(ad);
        _create_raw_text(ad);
        _create_image_box(ad);
        _create_data_list(ad);
        _create_buttons(ad);

        /*
         * Get the path to media
         */
        _path_to_media_init(ad);
    }
}

/*
 * Create popup window asking to enter application id
 */
static void _create_pop_win(appdata_s *ad)
{
    Evas_Object *pop_win = elm_win_inwin_add(ad->win);
    evas_object_show(pop_win);

    Evas_Object *pop_box = elm_box_add(pop_win);
    evas_object_size_hint_align_set(pop_box, EVAS_HINT_FILL, EVAS_HINT_FILL);
    evas_object_size_hint_weight_set(pop_box, EVAS_HINT_EXPAND, EVAS_HINT_EXPAND);
    elm_win_inwin_content_set(pop_win, pop_box);
    evas_object_show(pop_box);

    Evas_Object *label = elm_label_add(pop_box);
    elm_object_text_set(label, "Enter the API key from your openweathermap.org:");
    elm_box_pack_end(pop_box, label);
    evas_object_show(label);

    Evas_Object *app_id_entry = elm_entry_add(pop_box);
    elm_object_part_text_set(app_id_entry, "guide", "Enter the API key here");
    evas_object_size_hint_align_set(app_id_entry, EVAS_HINT_FILL, 0.0);
    evas_object_size_hint_weight_set(app_id_entry, EVAS_HINT_EXPAND, 0.0);
    elm_box_pack_end(pop_box, app_id_entry);
    evas_object_show(app_id_entry);

    static Elm_Entry_Filter_Limit_Size limit_size = {
        .max_char_count = APP_ID_LEN,
        .max_byte_count = 0
    };

    elm_entry_markup_filter_append(app_id_entry, elm_entry_filter_limit_size, &limit_size);

    Evas_Object *ok_button = elm_button_add(pop_box);
    elm_object_text_set(ok_button, "Ok");
    evas_object_size_hint_align_set(ok_button, EVAS_HINT_FILL, 0.0);
    evas_object_size_hint_weight_set(ok_button, EVAS_HINT_EXPAND, 0.0);
    evas_object_smart_callback_add(ok_button, "clicked", _pop_win_hide, ad);
    evas_object_data_set(ok_button, "entry", app_id_entry);
    evas_object_data_set(ok_button, "inwin", pop_win);
    elm_box_pack_end(pop_box, ok_button);
    evas_object_show(ok_button);

    label = elm_label_add(pop_box);
    elm_object_text_set(label, "<br><br>");
    elm_box_pack_end(pop_box, label);
    evas_object_show(label);

    Evas_Object *sign_up = elm_button_add(pop_box);
    elm_object_text_set(sign_up, "Sign up to get API key");
    evas_object_size_hint_align_set(sign_up, EVAS_HINT_FILL, 0.0);
    evas_object_size_hint_weight_set(sign_up, EVAS_HINT_EXPAND, 0.0);
    evas_object_smart_callback_add(sign_up, "clicked", _open_sign_up_site, ad);
    evas_object_data_set(sign_up, "inwin", pop_win);
    elm_box_pack_end(pop_box, sign_up);
    evas_object_show(sign_up);
}

static void _create_base_gui(appdata_s *ad)
{
    /* Window */
    ad->win = elm_win_util_standard_add(PACKAGE, PACKAGE);
    elm_win_conformant_set(ad->win, EINA_TRUE);
    elm_win_autodel_set(ad->win, EINA_TRUE);
    elm_win_indicator_mode_set(ad->win, ELM_WIN_INDICATOR_SHOW);
    elm_win_indicator_opacity_set(ad->win, ELM_WIN_INDICATOR_OPAQUE);

    evas_object_smart_callback_add(ad->win, "delete, request", win_delete_request_cb, NULL);

    /* Create conformant */
    ad->conform = elm_conformant_add(ad->win);
    evas_object_size_hint_weight_set(ad->conform, EVAS_HINT_EXPAND, EVAS_HINT_EXPAND);
    elm_win_resize_object_add(ad->win, ad->conform);
    evas_object_show(ad->conform);
    eext_object_event_callback_add(ad->win, EEXT_CALLBACK_BACK, win_back_cb, ad);

    ad->weather_info_list = NULL;

    _create_pop_win(ad);

    /* Show window after base gui is set up */
    evas_object_show(ad->win);
}

static bool storage_info_cb(int storage_id, storage_type_e type, storage_state_e state,
                            const char *path, void *user_data)
{
    int *id = (int *) user_data;
    if (type == STORAGE_TYPE_INTERNAL) {
        *id = storage_id;
        return false;
    }
    return true;
}

static int get_internal_storage_id(int *id)
{
    int error;
    int found_id;
    error = storage_foreach_device_supported(storage_info_cb, &found_id);
    if (error != STORAGE_ERROR_NONE) {
        dlog_print(DLOG_ERROR, LOG_TAG, "storage_foreach_device_supported() failed: %d", error);
        return error;
    }
    *id = found_id;
    return STORAGE_ERROR_NONE;
}

static bool app_create(void *data)
{
    appdata_s *ad = data;
    ad->internal_storage_id = -1;
    int error = get_internal_storage_id(&ad->internal_storage_id);
    if (error != STORAGE_ERROR_NONE) {
        dlog_print(DLOG_ERROR, LOG_TAG, "Could not get internal storage ID");
        return false;
    }

    ad->folder_path = NULL;

    /*
     * Prepare variables
     */
    _prepare_ui(ad);

    /*
     * Create UI
     */
    _create_base_gui(ad);

    /*
     * Get data from web
     */
    _get_xml_data(ad);

    return true;
}

static void app_terminate(void *data)
{
    appdata_s *ad = data;
    free(ad->folder_path);
    xmlCleanupParser();
    xmlMemoryDump();
}

int main(int argc, char *argv[])
{
    appdata_s ad;
    memset(&ad, 0x00, sizeof(appdata_s));

    ui_app_lifecycle_callback_s event_callback;
    memset(&event_callback, 0x00, sizeof(ui_app_lifecycle_callback_s));

    event_callback.create = app_create;
    event_callback.terminate = app_terminate;

    int ret = ui_app_main(argc, argv, &event_callback, &ad);
    if (ret != APP_ERROR_NONE)
        dlog_print(DLOG_ERROR, LOG_TAG, "> ui_app_main() failed with error %d", ret);

    return ret;
}
