/*
 * Copyright (c) 2015 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <download.h>
#include "main.h"
#include "ui.h"

const int CHBUF_SIZE = 200;
char chbuf[CHBUF_SIZE];

/*
 * Convert temperature units from Kelvin to Celcius
 */
static int k_to_c(char *kelvin)
{
    return atof(kelvin) - 273.15;
}

/*
 * Check if file exists
 */
static int file_exist(char *filename)
{
    struct stat data;
    return (stat(filename, &data) == 0);
}

/*
 * Extract specified data from XML nodes
 */
static void get_data_from_node(xmlNode *cur_node, char *chbuf, appdata_s *ad)
{
    xmlChar *tmp1, *tmp2, *tmp3;

    /*
     * !strcmp(x, y)==TRUE, when x==y
     */
    if (!strcmp((char *)cur_node->name, "city")) {
        /*
         * xmlGetProp() gets property specified with string from given XML node
         */
        snprintf(chbuf, CHBUF_SIZE, "%s: %s", cur_node->name, tmp1 =
                     xmlGetProp(cur_node, (const xmlChar *)"name"));
        xmlFree(tmp1);
    }

    else if (!strcmp((char *)cur_node->name, "temperature")) {
        snprintf(chbuf, CHBUF_SIZE, "temp: %s %s | %d\"C", tmp1 =
                     xmlGetProp(cur_node, (const xmlChar *)"value"), tmp2 =
                     xmlGetProp(cur_node, (const xmlChar *)"unit"),
                 k_to_c((char *)(tmp3 = xmlGetProp(cur_node, (const xmlChar *)"value"))));

        xmlFree(tmp1);
        xmlFree(tmp2);
        xmlFree(tmp3);
        /*
         * ad->temp is set from current node property
         * ad->humidity and ad->temp is used to set image in main window
         */
        ad->temp = k_to_c((char *)(tmp1 = xmlGetProp(cur_node, (const xmlChar *)"value")));
        xmlFree(tmp1);
    }

    else if (!strcmp((char *)cur_node->name, "coord")) {
        snprintf(chbuf, CHBUF_SIZE, "<tab> Lon: %s Lat: %s", tmp1 =
                     xmlGetProp(cur_node, (const xmlChar *)"lon"), tmp2 =
                     xmlGetProp(cur_node, (const xmlChar *)"lat"));
        xmlFree(tmp1);
        xmlFree(tmp2);
    }

    else if (!strcmp((char *)cur_node->name, "country")) {
        /*
         * xmlNodeGetContent() gets content from given node
         * ex. <tag>some data</tag> , xmlNodeGetContent() will return "some data"
         */
        snprintf(chbuf, CHBUF_SIZE, "<tab> %s", tmp1 = xmlNodeGetContent(cur_node));
        xmlFree(tmp1);
    }

    else if (!strcmp((char *)cur_node->name, "sun")) {
        snprintf(chbuf, CHBUF_SIZE, "<tab> %s set: %s", cur_node->name, tmp1 =
                     xmlGetProp(cur_node, (const xmlChar *)"set"));
        xmlFree(tmp1);
    }

    else if (!strcmp((char *)cur_node->name, "humidity")
             || !strcmp((char *)cur_node->name, "pressure")) {
        snprintf(chbuf, CHBUF_SIZE, "%s: %s%s", cur_node->name, tmp1 =
                     xmlGetProp(cur_node, (const xmlChar *)"value"), tmp2 =
                     xmlGetProp(cur_node, (const xmlChar *)"unit"));

        xmlFree(tmp1);
        xmlFree(tmp2);

        /*
         * ad->humidity is set from current node property
         * ad->humidity and ad->temp is used to set image in main window
         */
        if (!strcmp((char *)cur_node->name, "humidity")) {
            ad->humidity = atoi((char *)(tmp1 = xmlGetProp(cur_node, (const xmlChar *)"value")));
            xmlFree(tmp1);
        }
    }

    else if (!strcmp((char *)cur_node->name, "wind"))
        snprintf(chbuf, CHBUF_SIZE, "%s: ", cur_node->name);

    else if (!strcmp((char *)cur_node->name, "speed")) {
        snprintf(chbuf, CHBUF_SIZE, "<tab>%s", tmp1 =
                     xmlGetProp(cur_node, (const xmlChar *)"name"));
        xmlFree(tmp1);
    }

    else if (!strcmp((char *)cur_node->name, "direction")) {
        snprintf(chbuf, CHBUF_SIZE, "<tab>%s", tmp1 =
                     xmlGetProp(cur_node, (const xmlChar *)"name"));
        xmlFree(tmp1);
    }

    else if (!strcmp((char *)cur_node->name, "clouds")) {
        snprintf(chbuf, CHBUF_SIZE, "%s: %s", cur_node->name, tmp1 =
                     xmlGetProp(cur_node, (const xmlChar *)"name"));
        xmlFree(tmp1);
    }

    else {
        snprintf(chbuf, CHBUF_SIZE, "%s: %s %s", cur_node->name, tmp1 =
                     xmlNodeGetContent(cur_node), tmp2 =
                     xmlGetProp(cur_node, (const xmlChar *)"value"));
        xmlFree(tmp1);
        xmlFree(tmp2);
    }
}

/*
 * Recursively iterate thought XML tree
 */
static void iterate_element(xmlNode *a_node, appdata_s *ad)
{
    xmlNode *cur_node = NULL;

    for (cur_node = a_node; cur_node; cur_node = cur_node->next) {
        /*
         * Check if handle refers for node, not ex. node text
         */
        if (cur_node->type == XML_ELEMENT_NODE) {
            dlog_print(DLOG_DEBUG, LOG_TAG, "node type: Element, name: %s\n", cur_node->name);

            /*
             * Perform action on all nodes different from root ("current")
             */
            if (strcmp((char *)cur_node->name, "current")) {
                /*
                 * Prepare temporary buf for weather feature
                 */
                memset(chbuf, 0x00, CHBUF_SIZE);

                /*
                 * Get weather feature from node
                 */
                get_data_from_node(cur_node, chbuf, ad);
                chbuf[CHBUF_SIZE - 1] = 0x00;

                /*
                 * Duplicate data to eina_list for further displaying
                 */
                ad->weather_info_list = eina_list_append(ad->weather_info_list, strdup(chbuf));
            }
        }

        iterate_element(cur_node->children, ad);
    }
}

/*
 * Replace XML special sings, add Evas Entry formating tags
 */
static void strp(char *in, char *out)
{
    memcpy(out, "<code><string>", 14);
    out += 6;

    while (*in != 0x00) {
        if (*in != '<' && *in != '>')
            *out++ = *in++;
        else {
            *out++ = '&';

            if (*in == '<') {
                *out++ = 'l';
                *out++ = 't';
            } else {
                *out++ = 'g';
                *out++ = 't';
            }

            *out++ = ';';
            in++;
        }
    }

    memcpy(out, "</string></code>", 16);
}

/*
 * Count XML special characters
 */
static int counttags(char *s)
{
    int ret = 0;

    while (*s != '\0') {
        if (*s == '<' || *s == '>')
            ret++;

        s++;
    }

    dlog_print(DLOG_DEBUG, LOG_TAG, " >Found %d xml tags", ret);
    return ret;
}

/*
 * Convert XML to text that may be displayed in Evas Entry
 */
static void prepare_raw_text(appdata_s *ad)
{
    dlog_print(DLOG_DEBUG, LOG_TAG, "> prepare_raw_text");
    free(ad->xml_to_str);

    /*
     * Memory allocation for larger text
     * +30 is memory for <code><string></code></string>
     * *3 is memory for replacing < and >, with &gt; and &lt;
     */
    ad->xml_to_str =
        malloc(sizeof(char) * ((strlen((char *)ad->buf) + 3 * counttags((char *)ad->buf)) + 30));
    strp((char *)ad->buf, ad->xml_to_str);
    elm_object_text_set(ad->entry, ad->xml_to_str);
}

/*
 * Parse XML document, extract XML tree
 */
static void parse_xml(char *input, appdata_s *ad)
{
    xmlNode *root_element = NULL;

    dlog_print(DLOG_DEBUG, LOG_TAG, "parse_xml");

    /*
     * Extract XML tree
     */
    ad->doc = xmlParseDoc((const xmlChar *)input);

    if (ad->doc == NULL) {
        dlog_print(DLOG_ERROR, LOG_TAG, "failed to parse the including file\n");
        return;
    }

    root_element = xmlDocGetRootElement(ad->doc);

    /*
     * Start geting data from nodes
     */
    iterate_element(root_element, ad);

    /*
     * Set image in main window, basing on data from iterate_element()
     */
    _set_weather_image(ad->temp, ad->humidity, ad);
}

/*
 * Called moment after downloading
 */
static int download_id = -1;
static void parse_downloaded(void *data)
{
    appdata_s *ad = (appdata_s *)data;
    char *file_name = NULL;
    download_get_downloaded_file_path(download_id, &file_name);
    dlog_print(DLOG_DEBUG, LOG_TAG, "got file %s", file_name);

    /*
     * Prepare memory for new XML tree
     */
    xmlFreeDoc(ad->doc);

    FILE *fp;
    char *line = NULL;
    size_t len = 0;
    ssize_t read;

    fp = fopen(file_name, "r");

    if (fp == NULL) {
        dlog_print(DLOG_ERROR, LOG_TAG, "fopen failed");
        return;
    }

    char *buf_end = (char *)ad->buf;
    int buf_fill_len = 0;
    while ((read = getline(&line, &len, fp)) != -1) {
        if (buf_fill_len >= XML_BUF_SIZE)
            break;
        strncpy(buf_end, line, XML_BUF_SIZE - buf_fill_len);
        buf_fill_len += len;
        buf_end += len;
        free(line);
    }
    fclose(fp);

    /*
     * Parse XML
     */
    parse_xml((char *)ad->buf, ad);

    /*
     * Prepare text to display in popup box
     */
    prepare_raw_text(ad);

    /* Remove the file when no longer used */
    remove(file_name);

    Eina_List *tmp;
    void *list_data;

    /*
     * Create new display list with list of features provided by _get_xml_data()
     */
    EINA_LIST_FOREACH(ad->weather_info_list, tmp, list_data) {
        elm_list_item_append(ad->display_list, (char *)list_data, NULL, NULL, NULL, NULL);
        dlog_print(DLOG_DEBUG, LOG_TAG, "List item:%s\n", list_data);
    }

    /*
     * Show list
     */
    elm_list_go(ad->display_list);
}

/*
 * Checks state of download and calls parsing when downloaded
 */
static void state_changed_cb(int _download_id, download_state_e state, void *user_data)
{
    switch (state) {
    case DOWNLOAD_STATE_QUEUED:
    case DOWNLOAD_STATE_DOWNLOADING:
        dlog_print(DLOG_DEBUG, LOG_TAG, "Download in progress");
        break;

    case DOWNLOAD_STATE_COMPLETED:
        ;
        download_unset_state_changed_cb(download_id);
        download_id = _download_id;
        dlog_print(DLOG_DEBUG, LOG_TAG, "Download completed");
        ecore_job_add(parse_downloaded, (void *)user_data);
        break;

    default:
        dlog_print(DLOG_DEBUG, LOG_TAG, "Download failed");
        download_destroy(download_id);
        break;
    }
}

/*
 * Downloads file from given url
 */
static void open_url_data(char *url, appdata_s *ad)
{
    int download_id;
    int error = download_create(&download_id);

    if (error != 0) {
        dlog_print(DLOG_DEBUG, LOG_TAG, "Couldn't download");
        return;
    }

    error = download_set_state_changed_cb(download_id, state_changed_cb, ad);

    if (error != 0) {
        download_destroy(download_id);
        return;
    }

    error = download_set_url(download_id, url);

    if (error != 0) {
        download_destroy(download_id);
        return;
    }

    char *data_path = app_get_data_path();
    download_set_destination(download_id, data_path);
    free(data_path);

    error = download_set_file_name(download_id, "tmp.xml");

    if (error != 0) {
        download_destroy(download_id);
        return;
    }

    error = download_start(download_id);

    if (error != 0)
        download_destroy(download_id);
}

/*
 * Download XML file specified by user
 */
void _get_xml_data(appdata_s *ad)
{
    /*
     * Getting XML Stream
     */
    char urlbuf[500];

    /*
     * OpenWeatherMap service provides open weather data for more than 200, 000
     * cities and any geo location that is available on web-site and through API.
     */
    snprintf(urlbuf, 499, "http://api.openweathermap.org/data/2.5/weather?q=%s&mode=xml&APPID=%s",
             ad->city, ad->app_id);
    urlbuf[499] = '\0';

    dlog_print(DLOG_DEBUG, LOG_TAG, "> URL %s", urlbuf);

    /*
     * Open specified URL
     */
    open_url_data(urlbuf, ad);
}

/*
 * Create XML file, containing records chosen by user
 */
void _store_weather_to_file(appdata_s *ad)
{
    xmlChar *tmp;
    xmlDocPtr doc = NULL;
    xmlNodePtr root_node = NULL, node = NULL, add_node = NULL;
    xmlDtdPtr dtd = NULL;

    /*
     * Prepare XML tree
     */
    if (ad->doc == NULL)
        ad->doc = xmlParseDoc((xmlChar *)ad->buf);

    /*
     * Check if user already stored any data
     */
    char path[BUFLEN];
    snprintf(path, BUFLEN, "%s/%s", ad->folder_path, "xml.xml");
    dlog_print(DLOG_DEBUG, LOG_TAG, "File : %s exists", path);
    if (file_exist(path)) {
        /*
         * Open existing XML file, get tree and root node
         */
        dlog_print(DLOG_DEBUG, LOG_TAG, "File : %s exists", path);
        doc = xmlReadFile(path, NULL, 0);
        if (doc == NULL)
            dlog_print(DLOG_DEBUG, LOG_TAG, "error: could not parse file %s", path);
        root_node = xmlDocGetRootElement(doc);
    } else {
        /*
         * Create new XML tree, root node
         */
        dlog_print(DLOG_DEBUG, LOG_TAG, "Creating : %s", path);
        doc = xmlNewDoc(BAD_CAST "1.0");
        root_node = xmlNewNode(NULL, BAD_CAST "root");
        xmlDocSetRootElement(doc, root_node);
        dtd = xmlCreateIntSubset(doc, BAD_CAST "root", NULL, BAD_CAST "tree2.dtd");
    }

    /*
     * Prepare data save time
     */
    time_t rawtime;
    struct tm *timeinfo;

    time(&rawtime);
    timeinfo = localtime(&rawtime);

    /*
     * Add new node storing separated weather records, set creation date as a content
     */
    node = xmlNewChild(root_node, NULL, BAD_CAST "state", BAD_CAST asctime(timeinfo));

    /*
     * Copy tree obtained before from URI into local XML document / XML tree
     */
    add_node = xmlDocGetRootElement(ad->doc);

    if (add_node == NULL)
        return;

    xmlAddChild(node, add_node);

    /*
     * Set city name from child tree in node properties
     */
    xmlNodePtr cur = add_node->children;

    while (cur != NULL) {
        dlog_print(DLOG_DEBUG, LOG_TAG, " >node %s", cur->name);

        if ((!xmlStrcmp(cur->name, (const xmlChar *)"city"))) {
            dlog_print(DLOG_DEBUG, LOG_TAG, "city found");
            xmlNewProp(node, BAD_CAST "city", tmp = xmlGetProp(cur, (const xmlChar *)"name"));
            xmlFree(tmp);
        }

        cur = cur->next;
    }

    /*
     * Save tree to XML file
     */
    xmlSaveFormatFileEnc(path, doc, "UTF-8", 1);
    xmlFreeDoc(doc);
    ad->doc = NULL;
}
